/* total.c
 *
 * total - summarizes columnar data
 * By Jon Rifkin <jon.rifkin@uconn.edu>
 * Copyright 1999-2005 Jonathan Rifkin
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


/*
------------------------------------------------------------------------
History
------------------------------------------------------------------------
*/
/*
Jan 23, 2000
	Started
Oct 19, 2001
	Added Queue option
*/

/*

------------------------------------------------------------------------
Compile Switches
------------------------------------------------------------------------
*/
#define PARSE_COLUMN
#define USE_SORT
#define GET_DATA_COL


/*
------------------------------------------------------------------------
Include files
------------------------------------------------------------------------
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include "hash.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif



/*
------------------------------------------------------------------------
Usage
------------------------------------------------------------------------
*/
char *Usage_m[] = {
"",
"   total -dmqsvFN <key-col> <data-col> <file>",
"",
"     -d     Debug",
"     -f M   Print first M records (must also use sort option)",
"     -q N   Store at most N records, printing overflow as they're deleted",
"     -s I[r][,J[r][,...]]",
"            Sort records by columns I,J,..; in reverse order if 'r'",
"     -v     Print version info and exit",
"     -F c   Use character c as field delimiter",
"     -N H   Use H number of slots in hash table (-v option prints default)",
"",
"   Read text data <file>.  Each line contains a record consisting",
"   of space delimited fields, some of which are numeric.  Use",
"   <key-col> (comma delimited list of columns starting at 1, or '-'",
"   for no key) as record keys, and for each unique set of record",
"   keys calculate the column values requested in <data-col>.",
"",
"   <key-col>  comma delimited list of column numbers OR just '-'",
"              to indicate no keys, this groups all file records together.",
"   <data-col> comma delimited list of column numbers follow by",
"                   s  Sum (the default).",
"                   a  Average.",
"                   d  Standard deviation.",
"                   m  Minimum.",
"                   x  Maximum.",
"                   n  Number of rows.",
"                   f  Column value from first row.",
"                   l  Column value from last row.",
"                   z  If sum is non-zero, set value to 1.",
"",
"   Examples:",
"",
"     Input file:    burt dog  10 5",
"                    burt fish  1 1",
"                    bill dog   2 2",
"                    burt dog   5 5",
"                    bill fish  3 1",
"",
"     Command:       total 1,2 3,4 input.fil",
"",
"     Output:        burt dog  15 10",
"                    burt fish  1  1",
"                    bill dog   2  2",
"                    bill fish  3  1",
"",
"     Command:       total 2 4,3,n input.fil",
"",
"     Output:        dog  12 17 3",
"                    fish  2  4 2",
"",
"     Command:       total - n input.fil",
"                    5",
"",
VERSION,
0
};


						 
/*
------------------------------------------------------------------------
DEFINES
------------------------------------------------------------------------
*/
#define MAX_COL 256
#define NSTR    4192

#define U_CHAR unsigned char

/*  Number of slots in hash table  */
#define N_HASH_SLOTS  500000

/*  Constant to denote unknown action  */
#define FOUND_FALSE -1

#define FIRST_FALSE 0
#define FIRST_TRUE  1



/*
------------------------------------------------------------------------
Debugging Macros
------------------------------------------------------------------------
*/
#define WRITETXT(TXT) \
	printf ("FILE %s LINE %i: \"%s\"\n", __FILE__, __LINE__, TXT); \
	fflush (stdin);

#define WRITEMSG \
	printf ("In file %s at line %i.\n", __FILE__, __LINE__); \
	fflush (stdin);

#define WRITEVAR(VAR_NAME,VAR_TYPE) \
	printf ("FILE %s LINE %i: ", __FILE__, __LINE__); \
	printf ("%s <", #VAR_NAME); \
	printf (#VAR_TYPE, (VAR_NAME) ); \
	printf (">\n"); \
	fflush (stdin); 


/*
------------------------------------------------------------------------
Extern variables
------------------------------------------------------------------------
*/
/*  used by getop() function  */
extern char *optarg;

/*
------------------------------------------------------------------------
Types
------------------------------------------------------------------------
*/

/*   This structure holds several kinds of information.  In its most 
 *   general use it describes a transformation of one array to another.
 *   Elements of src[] and dst[] define positions an input and
 *   output array, while action[] is a character used to identify an
 *   operation maps src[] to dst[].
 *
 *   It can also use to descibe a list of columns to sort on (in which
 *   case only src[] is used).  */
typedef struct {
	int src[MAX_COL];
	int dst[MAX_COL];
	int action[MAX_COL];
	int n;
} action_t;

/*
------------------------------------------------------------------------
File variables
------------------------------------------------------------------------
*/
char *comment_char_m = "#";
int  debug_m = 0;
action_t sort_action_m;

/*  Elements data_col_m[i], data_action_m[i] store the i'th
 *  output column specifictions entered by the user.
 */
action_t data_action_m;

/*  Elements proc_col_m[i], proc_action_ms[i] store a unique
 *  combination of input data column and action (sum, min, max, etc)
 *  on that column.  A input single column n might appear in two
 *  different elements of proc_col_m[] if there are two actions done
 *  on that input column.
 */
action_t proc_action_m;
action_t post_action_m;
int perform_post_m = 0;


/*
------------------------------------------------------------------------
Local Function Prototypes
------------------------------------------------------------------------
*/
int is_whitespace(int);
int is_comment(char *);
int find_action_col (action_t *, int, int);
int get_fields (char *s, char **fptr, int max_field, int field_sep);
void prformat(double);
int isnumber (char *);
void print_key    (helem_t  *);
void walk_history (htable_t *ht);
int dcmp (const void *aptr, const void *bptr);
double get_data_col ( helem_t *, int);
int get_max_index ( helem_t **, int, int (*cmp)(const void *, const void *));
void apply_action(double *dout, char **fptr, action_t *proc, int firsttime);
void parse_column       (char *, int, action_t *);
void get_proc_from_data (action_t *,   action_t *);

/*  Find maximum position stored in src and dst arrays in action_t  */
int  get_max_src (action_t *, int);
int  get_max_dst (action_t *, int);

/*  Make key  */
void make_key(char *keybuff, char **fptr, action_t *key, int field_sep, int nlen);
void Usage(void);

void debug_action(char *msg, action_t *a);

/*
------------------------------------------------------------------------
Main function
------------------------------------------------------------------------
*/
int main (int argc, char *argv[]) {
	int nfield;
	action_t key_action;
	int i;
	char *keystr, *datastr;
	char buffer [NSTR];
	int    nfirstdata;
	double *firstdata = NULL;
	double *rdata   = NULL;
	char keybuff[NSTR];
	char *fptr[MAX_COL];
	htable_t *ht = NULL;
	helem_t  *t        = NULL;
	FILE *fin;
	int max_col;
	int optchar;
	int nhashslots = N_HASH_SLOTS;
	int max_queue_length=0;
	int max_output=0;
	int use_sort  =0;
	int use_top_sort = 0;
	helem_t **index = NULL;
	int field_sep = ' ';
	int max_index;

	/*  Read command line options  */
	while (-1 != (optchar=getopt(argc,argv,"F:N:df:q:s:v"))) {
		switch (optchar) {
			case '?':
			return 1;
		/*  Debugging option  */
		case 'd':
			debug_m = 1;
			printf ("Debugging mode is on\n");
			break;
		case 'v':
			printf ("Version %s (compiled %s)\n", VERSION,__DATE__);
			printf ("Default number of hash slots = %d\n", N_HASH_SLOTS);
			return 0;
			break;
		case 'N':
			nhashslots = atoi(optarg);
			if (nhashslots<1) {
				printf (
				"ERROR: Number of hash slots (-N%d) must be one or greater\n", 
				nhashslots);
				exit(1);
			}
			break;
		case 'F':
			field_sep = optarg[0];
			break;
		case 'q':
			max_queue_length = atoi(optarg);
			if (max_queue_length<0) max_queue_length=0;
			break;
		case 'f':
			max_output = atoi(optarg);
			break;
		case 's':
			use_sort   = 1;
			parse_column (optarg, 'f', &sort_action_m);
debug_action("sort", &sort_action_m);
			break;
		default:
			return 1;
		}
	}

	/*  Print Usage  */
	if (argc<4) {
		Usage();
		return 0;
	}
	
	/*  Parse key column argument  */
	keystr = strdup (argv[optind++]);

	parse_column (keystr, 's', &key_action);
debug_action("key_action", &key_action);

	/*  Parse output data column argument which is a comma separated
	 *  string of numbers followed by an optional letter, for example
	 *  "3,4,5a".  The numbers and letters are stored in separate arrays
	 *  data_col_m[] and data_action_m[] in the order with which they are
	 *  read in.  If no letter is present the default 's' is used.  */
	datastr = strdup (argv[optind++]);
	parse_column (datastr, 's', &data_action_m);
debug_action("data_action", &data_action_m);

	get_proc_from_data (&data_action_m, &proc_action_m);
debug_action("proc_action", &proc_action_m);

	post_action_m.n = 0;
	post_action_m.src[0] = 0;
	post_action_m.dst[0] = 0;
	post_action_m.action[0] = 'n';
	/*  We might not need to perform a post calculation  */
	// perform_post_m = 0;


	/*  Open file  (use stdin if file "-")  */
	if (!strcmp("-",argv[optind])) {
		fin = stdin;
	} else {
		fin = fopen (argv[optind], "rt");
		if (NULL==fin) {
			printf ("ERROR:  Cannot open input file <%s>\n", argv[optind]);
			return 1;
		}
	}

	/*  Find maximum column that must be read from each input line.  This
	 *  is determine from the keys columns and data columns requested.
	 *  By knowing the max column, and if the max column needed is less
	 *  than the total number of columns in the input line, we can save
	 *  significant work by spliting only the needed number of fields
	 *  instead of spliting all the fields.  */
	max_col = -1;
debug_action("key_action", &key_action);
	max_col = get_max_src(&key_action, max_col);
debug_action("data_action", &data_action_m);
	max_col = get_max_src(&data_action_m, max_col);
	
	/*  Storage for data to read when adding a new key to hash  */
	nfirstdata = get_max_dst(&proc_action_m,0)+1;
	firstdata  = (double *) calloc (nfirstdata, sizeof(double));

	/*  Initialize hash table  */
	ht = ht_init(nhashslots, max_queue_length ? HT_HISTORY : 0);

	/*  Read file  */
	while (NULL!=fgets(buffer,NSTR,fin)) {

		/*  Remove trailing new line  */
		buffer[strlen(buffer)-1] = '\0';

		/*  Skip comments  */
		if (is_comment(buffer))  continue;

		/*  Read fields from input line  */
		nfield = get_fields (buffer, fptr, max_col+1, field_sep);
		if (max_col>=nfield) continue;

		/*  Form key from input columns  */
		make_key(keybuff, fptr, &key_action, field_sep, NSTR);


		/*  Lookup key, process entry  */
		if ( 
			(t=ht_findelem (ht,(U_CHAR *)&keybuff ,strlen(keybuff)+1))
			) {

			/*  Get pointer to data for this hash element  */
			rdata = (double *) t->data;

			apply_action(rdata, fptr, &proc_action_m, FIRST_FALSE);
	
			/*  Make current key the newest, so data output can be newest
			 *  first  */
			ht_makenewest (ht, t);
			
		/*  No key found, create new entry  */
		} else {

			apply_action(firstdata, fptr, &proc_action_m, FIRST_TRUE);

			/*  Store new data with new key  */
			ht_storekey (ht, (U_CHAR *) &keybuff, strlen(keybuff)+1, 
				(U_CHAR *) firstdata, nfirstdata*sizeof(firstdata[0]));

		}  /* add new entry  */

		/*  Too many keys - print and remove oldest  */
		if (max_queue_length && ht_getcount(ht)>max_queue_length) {
			t = ht_getoldest(ht);
			print_key (t);
			ht_freeelem(ht,t);
		}

	}   /*  read file  */
		
	fclose(fin);

	/*  If no elements, then suppress sort  */
	if (ht_getcount(ht)==0) {
		use_sort = 0;
		max_output = 0;
	}

	/*  Post process the values.  Some user output requests require that
	 *  we perform further calcuations on the sums in the stored key,data
	 *  pairs before printing them.  Compuations will be made from the
	 *  stored data to a temporary data array, then we will copy the
	 *  values from the temporary array back to the stored data.  */
	if (perform_post_m) {
		while ( (t = ht_getnext(ht)) ) {
			rdata = (double *) t->data;
			for (i=0;i<post_action_m.n;i++) {
			}
		}
	}
	

	/*  When only printing the first N elements, a full sort is
	 *  relatively inefficient.  Instead, use a simple bubble sort if
	 *  first N element is less then 1/3 the total number.
	 *  This value of 1/3 is a very rough efficiany estimate, feel free
	 *  to change this if you have a better idea.
	 */
	if (use_sort && max_output && max_output<ht_getcount(ht)/3)
		use_top_sort = 1;

	/*  Print results in sorted order of data, top max_output only  */
	if (use_top_sort) {

		/*  Allocate index storage for number of top data  */
		index = (helem_t **) calloc(max_output, sizeof(helem_t *));
		if (NULL==index) {
			printf ("ERROR:  Cannot allocate memory for top_sort index\n");
			return 1;
		}

		/*  Place the first max_output elements into index  */
		ht_initwalk(ht);
		for (i=0;i<max_output;i++) {
			index[i] = ht_getnext(ht);
		}

		/*  Read through rest of list comparing with max_index  */
		max_index = get_max_index(index,max_output,dcmp);
		while ( (t = ht_getnext(ht)) ) {
			if (dcmp(index+max_index,&t)>0) {
				index[max_index] = t;
				max_index = get_max_index(index,max_output,dcmp);
			}
		}

		/*  Now have pointers to top max_output elements 
		 *  next sort and print
		 */
		qsort (index, max_output, sizeof(helem_t*), dcmp);
		for (i=0;i<max_output;i++) {
			t = index[i];
			print_key (t);
		}
		free (index);


	/*  Perform a full sort  */
	} else if (use_sort) {
		ht_initwalk(ht);
		index = (helem_t **) calloc (ht_getcount(ht), sizeof(helem_t *));
		if (NULL==index) {
			printf ("ERROR:  Cannot allocate memory for sort index\n");
			return 1;
		}
		i=ht_getcount(ht);
		do {
			i--;
			index[i] = ht_getnext(ht);
		} while (i);
		qsort (index, ht_getcount(ht), sizeof(helem_t*), dcmp);
		for (i=0;i<ht_getcount(ht);i++) {
			t = index[i];
			if (max_output && i>=max_output)
				break;
			print_key (t);
		}
		free (index);

	/*  Print results without sorting  */
	} else {
	
		/*  Write results  */
		ht_initwalk (ht);
		while ((t=ht_getnext(ht))) {
			/*  Print key and data  */
			print_key (t);
		}
	
	}

	/*  Print debug information if asked  */
	if (debug_m) ht_debuginfo(ht);
		
	/*  Don't bother freeing table, it takes alot of time cause we free each node individually,
	 *  just let OS free it up, it goes *much* faster
	 */
	/*  Free table  */
	/*
	ht_free (ht);
	*/

	return 0;
}
	


/*
------------------------------------------------------------------------
Local functions
------------------------------------------------------------------------
*/
int is_delimiter(int c, char *delimiter) {
	return (NULL!=strchr(delimiter,c));
}

int is_whitespace(int c) {
	return (NULL!=strchr(" \t\n\r",c));
}

int is_comment(char *s) {
	while (*s && is_whitespace(*s)) s++;
	return (*s==comment_char_m[0]);
}

/*  Return array of pointers fptr to fields in string  */
int get_fields (char *s, char **fptr, int max_field, int field_sep) {
	int nfield=0;

	/*  Fields separated by one or more blanks - leading blanks ignored  */
	if (field_sep==' ') {

		/*  Skip leading whitespace  */
		while (*s==' ' || *s=='\t') s++;
		/*  Find tokens  */
		while (*s && nfield<max_field) {
			fptr[nfield++] = s;
			/*  Find first whitespace  */
			while (*s!=' ' && *s && *s!='\t')  s++;
			/*  Make first white space string terminator */
			if (*s)  *s++ = '\0';
			/*  Find last whitespace */
			while (*s==' ' || *s=='\t')  s++;
		}

	/*  Fields separated by single field separator  */
	} else {
		
		/*  Find tokens  */
		while (*s && nfield<max_field) {
			fptr[nfield++] = s;
			/*  Find next field separator  */
			while (*s && *s!=field_sep)  s++;
			/*  Make field separator string terminator */
			if (*s)  *s++ = '\0';
		}

	}

	return nfield;
}


/*  Return list_col[] index which is result of 'action' on
 *  output data column 'col'
 */ 
int find_action_col (action_t *proc, int col, int action) {
	int i;
	for (i=0;i<proc->n;i++) {
		if (proc->src[i]==col && proc->action[i]==action) 
			return i;
	}
	return FOUND_FALSE;
}


/*  Print double in integer format if integer  */
void prformat(double d) {
	if (rint(d)==d)  printf (" %.0f", d);
	else             printf (" %e",   d);
}


/*  Return 0,1 if string non-numeric,numeric */
int isnumber (char *str) {
	while (*str) {
		if (*str<'0'  || *str>'9')
			return 0;
		str++;
	}
	return 1;
}


/*  Print key and data for this hash element  */
void print_key (helem_t *t) {
	int    idata;
	double val;

	/*  No element - print blank line  */
	if (t==NULL) {
		printf ("\n");
		return;
	}

	/*  Print key  */
	printf ("%s", t->key);
	/*  Print requested fields  */
	for (idata=0;idata<data_action_m.n;idata++) {
		val = get_data_col(t,idata);
		prformat (val);
	}
	printf ("\n");
}


/*  Given hash element and output data column description
 *  return output double value
 */
double get_data_col (
helem_t *t, 
int idata) {
	double *rdata = (double *) t->data;
	double avg;
	int    isum;
	int    iproc;

	/*  If data action is 'n', just need first element  */
	if (data_action_m.action[idata]=='n' || data_action_m.action[idata]=='-') {
		return rdata[0];
	}

	/*  Other actions  */
	iproc = find_action_col (&proc_action_m, data_action_m.src[idata], data_action_m.action[idata]);
	switch (data_action_m.action[idata]) {
		/*  Print sum, min, max as is  */
		case 's':
		case 'm':
		case 'x':
		case 'f':
		case 'l':
			return rdata[iproc];
			break;
		/*  Calculate average  */
		case 'a':
			if (rdata[0]==0) return 0.0;
			else             return rdata[iproc]/rdata[0];
			break;
		/*  Calculate standard deviation or error (std.dev./sqrt(n))  */
		case 'd':
		case 'e':
			if (rdata[0]==0) return 0.0;
			else {
				/*  Find corresponding sum value  */
				isum = find_action_col (&proc_action_m, data_action_m.src[idata], 's');
				if (isum==-1) {
					fprintf (stderr, "INTERNAL ERROR:  No sum stored for");
					fprintf (stderr, " standard deviation calculation\n");
					exit(2);
				}
				avg = rdata[isum]/rdata[0];
				if (data_action_m.action[idata] == 'e')  {
					return sqrt((rdata[iproc]/rdata[0]-avg*avg)/rdata[0]);
				}
				else
					return sqrt( rdata[iproc]/rdata[0]-avg*avg          );
			}
			break;
		case 'z':
			if (rdata[iproc]==0) return 0;
			else						return 1;
	}  /* end of data_action_m[] switch  */

	/*  Unknown action - treat it as n  */
	return rdata[0];
}


void walk_history (htable_t *ht) {
		  /*test*/
		  helem_t *e = ht_getoldest(ht);
		  while (e) {
			  print_key (e);
			  e = (helem_t *) e->newer;
		  }
}



/*  Callback function called by qsort() to order output records  */
int dcmp (const void *aptr, const void *bptr) {
	helem_t *a = * ( (helem_t **) aptr );
	helem_t *b = * ( (helem_t **) bptr );
	double diff;
	int i,j,idiff;

	for (i=0;i<sort_action_m.n;i++) {
		j = sort_action_m.src[i];
		diff = get_data_col(a,j) - get_data_col(b,j);
		if (diff==0.0) 
			continue;
		idiff = diff>0 ? 1 : -1;
		if (sort_action_m.action[i]=='r')
			return -idiff;
		else
			return idiff;
	}
	return 0;
}


/*  
Parse data column argument 

arg               string being parsed for data column and actions
data_col_m[]      column indices returned
data_action_m[]   action codes (a,m,n,s,x) returned
ndata             number of index,action pairs returned
def_char          default action code

Note that legal action code 'n' will result in column number -1.
Need to take precautions.

*/
void parse_column (char *arg, int def_char, action_t *action) {
	char *ptr;
	char lastchr;
	int  col;
	action->n=0;
	ptr=strtok(arg,",");
	while (ptr) {
		/*  Print error if too many data columns  */
		if (action->n>MAX_COL) {
			fprintf (stderr,
				"ERROR:  To many data columns specified.  Max allowed is %d\n", 
				MAX_COL);
			exit (1);
		}
		/*  Get trailing character if a,m,n,s,x  */
		lastchr = ptr[strlen(ptr)-1];
		if (!strchr("0123456789", lastchr)) {
			action->action[action->n] = lastchr;
			ptr[strlen(ptr)-1] = '\0';
		} else {
			action->action[action->n] = def_char;
		}
		/*  Read column number for this data item  */
		col = atoi(ptr)-1;
		action->src[(action->n)  ] = col;
		action->dst[(action->n)++] = col;
		ptr=strtok(NULL,",");
	}
}


/*  Search list of pointers for smallest element  */
int get_max_index (
		helem_t **index, 
		int n,
		int (*cmp)(const void *, const void *))
{
	int max = 0;
	int j;
	for (j=1;j<n;j++) {
		if ( cmp( index+max, index+j )<0 ) {
			max = j;
		}
	}
	return max;
}


/*  Apply action and input value to strored value  */
void apply_action(double *dout, char **fptr, action_t *proc, int firsttime) {
	int iproc;
	double *dptr, val;
	for (iproc=0;iproc<proc->n;iproc++) {
		if (proc->action[iproc]!='n') {
			val  = atof(fptr[proc->src[iproc]]);
		}
		dptr = dout + proc->dst[iproc];
		switch (proc->action[iproc]) {
			/*  Store sum  */
			case 'a':
			case 's':
			case 'z':
				if (firsttime) {
					*dptr  = val;
				} else {
					*dptr += val;
				}
				break;
			/*  Store variance (sum of squares)  */
			case 'd':
			case 'e':
				if (firsttime) {
					*dptr  = val*val; } else {
					*dptr += val*val;
				}
				break;
			/*  Do nothing, this is handled automatically  */
			case 'n':
				if (firsttime) {
					*dptr = 1.0;
				} else {
					*dptr += 1.0;
				}
				break;
			/*  Store minimum value  */
			case 'm':
				if (firsttime) {
					*dptr = val;
				} else {
					if (val<*dptr) *dptr = val;
				}
				break;
			/*  Store maximum value  */
			case 'x':
				if (firsttime) {
					*dptr = val;
				} else {
					if (val>*dptr) *dptr = val;
				}
				break;
			/*  Store first occurance  */
			case 'f':
				if (firsttime) *dptr = val;
				break;
			/*  Store last occurance  */
			case 'l':
				*dptr = val;
				break;
		}
	}
}


void get_proc_from_data (action_t *data, action_t *proc) {
	int i;


	proc->n = 0;
	proc->src   [ proc->n   ] =   0;
	proc->dst   [ proc->n   ] =   0;
	proc->action[(proc->n)++] = 'n';
	for (i=0;i<data->n;i++) {

		/*  See if combination of data_col[i],data_action[i] is not
		 *  already present in proc->src,proc->action.  If so, then we
		 *  don't need to store it again.  This can happen because some
		 *  data_action[] values (for example 'd' or 'e') will also
		 *  add an 's' value.  */
		if (FOUND_FALSE != find_action_col(proc, data->src[i], data->action[i]) )
			continue;

		/*  Print error if too many data items */
		if (proc->n==MAX_COL)  {
			fprintf (stderr, 
				"ERROR:  To many data actions specified.  Max allowed is %d\n", MAX_COL);
			exit (1);
		}

		/*  Save data column action in process  */
		switch (data->action[i]) {
			/*  Field counts happens automatically  */
			case 'n':
				break;
			/*  Standard deviation requires summation also  */
			case 'd':
			case 'e':
				proc->action[proc->n] = data->action[i];
				proc->src   [proc->n] = data->src[i];
				proc->dst   [proc->n] = proc->n;
				proc->n++;
				/*  Include summation for this column if not already present  */
				if (FOUND_FALSE == find_action_col (proc, data->src[i], 's')) {
					proc->action[proc->n] = data->action[i];
					proc->src   [proc->n] = data->src[i];
					proc->dst   [proc->n] = proc->n;
					proc->n++;
				}
				break;
			/*  'Simple' processes */
			case 'm':  /*  Minimum  */
			case 'x':  /*  Maximum  */
			case 'a':  /*  Average  */
			case 's':  /*  Sum      */
			case 'f':  /*  First    */
			case 'l':  /*  Last     */
			case 'z':  /*  Zero/One */
				proc->action[proc->n] = data->action[i];
				proc->src   [proc->n] = data->src[i];
				proc->dst   [proc->n] = proc->n;
				proc->n++;
				break;
		}
	}
}

int get_max_src(action_t *a, int max) {
	int i;
	for (i=0;i<a->n;i++)
		if (max < a->src[i])  max = a->src[i];
	return max;
}

int get_max_dst(action_t *a, int max) {
	int i;
	for (i=0;i<a->n;i++)
		if (max < a->dst[i])  max = a->dst[i];
	return max;
}

//		make_key(keybuff, fptr, &key_action, field_sep, NSTR);
void make_key(char *keybuff, char **fptr, action_t *key, int field_sep, int nlen) {
	int ikey;
	char sep_str[2] = "\0\0";
	sep_str[0] = field_sep;
	keybuff[0] = '\0';
	nlen--;
debug_action("make_key: key:", key);
	if (key->n) {
		for (ikey=0;ikey<key->n;ikey++) {
			strncat (keybuff, fptr[key->src[ikey]],nlen);
			nlen -= strlen(fptr[key->src[ikey]]);
			strncat (keybuff, sep_str, nlen);
			nlen -= 1;
		}
	}
}

void debug_action(char *msg, action_t *a) {
	int i;
	printf ("actionname action_p len: (%s) (%p) (%d)\n", msg, a, a->n);
	for (i=0;i<a->n;i++) 
		printf ("src dst action %2d %2d %c\n", a->src[i], a->dst[i], a->action[i]);
}

void Usage(void) {
		char **ptr = Usage_m;
		printf ("%s\n", *ptr);
		while (*ptr) {
			printf ("%s\n", *(ptr++));
		}
}

