/* total.c
 *
 * total - summarizes columnar data
 * By Jon Rifkin <jon.rifkin@uconn.edu>
 * Copyright 1999-2005 Jonathan Rifkin
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


/*
------------------------------------------------------------------------
History
------------------------------------------------------------------------
*/
/*
Jan 23, 2000
	Started
Oct 19, 2001
	Added Queue option
2005-11-03
   Went crazy refactoring the entire program.
*/

/*

------------------------------------------------------------------------
Compile Switches
------------------------------------------------------------------------
*/
#define PARSE_COLUMN
#define USE_SORT
#define GET_DATA_COL


/*
------------------------------------------------------------------------
Include files
------------------------------------------------------------------------
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include "hash.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif



/*
------------------------------------------------------------------------
Usage
------------------------------------------------------------------------
*/
char *Usage_m[] = {
"",
"   total -dmqsvFN <key-col> <data-col> <file>",
"",
"     -d     Debug",
"     -f M   Print first M records (must also use sort option)",
"     -q N   Store at most N records, printing overflow as they're deleted",
"     -s I[r][,J[r][,...]]",
"            Sort records by columns I,J,..; in reverse order if 'r'",
"     -v     Print version info and exit",
"     -F c   Use character c as field delimiter",
"     -N H   Use H number of slots in hash table (-v option prints default)",
"",
"   Read text data <file>.  Each line contains a record consisting",
"   of space delimited fields, some of which are numeric.  Use",
"   <key-col> (comma delimited list of columns starting at 1, or '-'",
"   for no key) as record keys, and for each unique set of record",
"   keys calculate the column values requested in <data-col>.",
"",
"   <key-col>  comma delimited list of column numbers OR just '-'",
"              to indicate no keys, this groups all file records together.",
"   <data-col> comma delimited list of column numbers follow by",
"                   s  Sum (the default).",
"                   a  Average.",
"                   d  Standard deviation.",
"                   m  Minimum.",
"                   x  Maximum.",
"                   n  Number of rows.",
"                   f  Column value from first row.",
"                   l  Column value from last row.",
"                   z  If sum is non-zero, set value to 1.",
"",
"   Examples:",
"",
"     Input file:    burt dog  10 5",
"                    burt fish  1 1",
"                    bill dog   2 2",
"                    burt dog   5 5",
"                    bill fish  3 1",
"",
"     Command:       total 1,2 3,4 input.fil",
"",
"     Output:        burt dog  15 10",
"                    burt fish  1  1",
"                    bill dog   2  2",
"                    bill fish  3  1",
"",
"     Command:       total 2 4,3,n input.fil",
"",
"     Output:        dog  12 17 3",
"                    fish  2  4 2",
"",
"     Command:       total - n input.fil",
"                    5",
"",
VERSION,
0
};


						 
/*
------------------------------------------------------------------------
DEFINES
------------------------------------------------------------------------
*/
#define MAX_COL 256
#define NSTR    4192

#define U_CHAR unsigned char

/*  Number of slots in hash table  */
#define N_HASH_SLOTS  500000

/*  Constant to denote unknown action  */
#define FOUND_FALSE -1

#define FIRST_FALSE 0
#define FIRST_TRUE  1



/*
------------------------------------------------------------------------
Debugging Macros
------------------------------------------------------------------------
*/
#define WRITETXT(TXT) \
	printf ("FILE %s LINE %i: \"%s\"\n", __FILE__, __LINE__, TXT); \
	fflush (stdin);

#define WRITEMSG \
	printf ("In file %s at line %i.\n", __FILE__, __LINE__); \
	fflush (stdin);

#define WRITEVAR(VAR_NAME,VAR_TYPE) \
	printf ("FILE %s LINE %i: ", __FILE__, __LINE__); \
	printf ("%s <", #VAR_NAME); \
	printf (#VAR_TYPE, (VAR_NAME) ); \
	printf (">\n"); \
	fflush (stdin); 


/*
------------------------------------------------------------------------
Extern variables
------------------------------------------------------------------------
*/
/*  used by getop() function  */
extern char *optarg;

/*
------------------------------------------------------------------------
Types
------------------------------------------------------------------------
*/

/*   This structure holds several kinds of information.  In its most 
 *   general use it describes a transformation of one array to another.
 *   Elements of src[] and dst[] define positions an input and
 *   output array, while action[] is a character used to identify an
 *   operation maps src[] to dst[].
 *
 *   It can also use to descibe a list of columns to sort on (in which
 *   case only src[] is used).  */
typedef struct {
	double buf[MAX_COL];
	int src[MAX_COL];
	int dst[MAX_COL];
	int action[MAX_COL];
	int n;
	int nout;
} action_t;

/*
------------------------------------------------------------------------
File variables
------------------------------------------------------------------------
*/
char *comment_char_m = "#";
int  debug_m = 0;
action_t sort_action_m;

/*  Elements data_col_m[i], data_action_out_m[i] store the i'th
 *  output column specifictions entered by the user.
 */
action_t data_action_out_m;

/*  Elements proc_col_m[i], data_action_inc_ms[i] store a unique
 *  combination of input data column and action (sum, min, max, etc)
 *  on that column.  A input single column n might appear in two
 *  different elements of proc_col_m[] if there are two actions done
 *  on that input column.
 */
action_t data_action_inc_m;


/*
------------------------------------------------------------------------
Local Function Prototypes
------------------------------------------------------------------------
*/
int is_whitespace(int);
int is_comment(char *);
int find_action_col (action_t *, int, int);
int get_fields (char *s, char **fptr, int max_field, int field_sep);
void prformat(double);
int isnumber (char *);
void print_key    (helem_t  *);
void walk_history (htable_t *ht);
int dcmp (const void *aptr, const void *bptr);
int get_max_index ( helem_t **, int, int (*cmp)(const void *, const void *));
void parse_column       (char *, int, action_t *);
void get_inc_from_out (action_t *,   action_t *);

/*  Find maximum position stored in src and dst arrays in action_t  */
int  get_max_src (action_t *, int);
int  get_max_dst (action_t *, int);

/*  Make key  */
void make_key(char *keybuff, char **fptr, action_t *key, int field_sep, int nlen);
void Usage(void);

/*  Convert data to be stored  */
void convert_data_inc(action_t *proc, double *data, char **fptr, int firsttime);

/*  Convert data before it is written out  */
void convert_data_out(action_t *proc, double *data);

int  copychars(char *p, char *buf, int nbuf, int is_digits);
void parse_manual (char *arg, int def_char, action_t *t);

void debug_action(char *msg, action_t *a);

/*
------------------------------------------------------------------------
Main function
------------------------------------------------------------------------
*/
int main (int argc, char *argv[]) {
	int nfield;
	action_t key_action;
	int i;
	char *keystr, *datastr;
	char buffer [NSTR];
	int    nfirstdata;
	double *firstdata = NULL;
	double *rdata   = NULL;
	char keybuff[NSTR];
	char *fptr[MAX_COL];
	htable_t *ht = NULL;
	helem_t  *t        = NULL;
	FILE *fin;
	int max_col;
	int optchar;
	int nhashslots = N_HASH_SLOTS;
	int max_queue_length=0;
	int max_output=0;
	int use_sort  =0;
	int use_top_sort = 0;
	int use_manual   = 0;
	helem_t **index = NULL;
	int field_sep = ' ';
	int max_index;

	/*  Read command line options  */
	while (-1 != (optchar=getopt(argc,argv,"F:N:df:gmpq:s:v"))) {
		switch (optchar) {
			case '?':
			return 1;
		/*  Debugging option  */
		case 'd':
			debug_m = 1;
			printf ("Debugging mode is on\n");
			break;
		case 'v':
			printf ("Version %s (compiled %s)\n", VERSION,__DATE__);
			printf ("Default number of hash slots = %d\n", N_HASH_SLOTS);
			return 0;
			break;
		case 'N':
			nhashslots = atoi(optarg);
			if (nhashslots<1) {
				printf (
				"ERROR: Number of hash slots (-N%d) must be one or greater\n", 
				nhashslots);
				exit(1);
			}
			break;
		case 'F':
			field_sep = optarg[0];
			break;
		case 'q':
			max_queue_length = atoi(optarg);
			if (max_queue_length<0) max_queue_length=0;
			break;
		case 'f':
			max_output = atoi(optarg);
			break;
		case 's':
			use_sort   = 1;
			parse_column (optarg, 'f', &sort_action_m);
			break;
		case 'm':
			use_manual = 1;
			break;
		default:
			printf ("option (%c) not yet implemented\n", optchar);
			return 1;
		}
	}

	/*  Print Usage  */
	if ( (use_manual && argc<5) || argc<4) {
		Usage();
		return 0;
	}
	
	/*  Parse key column argument  */
	keystr = strdup (argv[optind++]);

	parse_column (keystr, 's', &key_action);

	/*  Parse output data column argument which is a comma separated
	 *  string of numbers followed by an optional letter, for example
	 *  "3,4,5a".  The numbers and letters are stored in separate arrays
	 *  data_col_m[] and data_action_out_m[] in the order with which they are
	 *  read in.  If no letter is present the default 's' is used.  */

	if (use_manual) {
		datastr = strdup (argv[optind++]);
		parse_manual (datastr, 's', &data_action_inc_m);
		datastr = strdup (argv[optind++]);
		parse_manual (datastr, 's', &data_action_out_m);

	} else {
		datastr = strdup (argv[optind++]);
		parse_column (datastr, 's', &data_action_out_m);

		get_inc_from_out (&data_action_out_m, &data_action_inc_m);
	}
#if 0
debug_action ("inc after:", &data_action_inc_m);
debug_action ("out after:", &data_action_out_m);
#endif

	/*  Open file  (use stdin if file "-")  */
	if (!strcmp("-",argv[optind])) {
		fin = stdin;
	} else {
		fin = fopen (argv[optind], "rt");
		if (NULL==fin) {
			printf ("ERROR:  Cannot open input file <%s>\n", argv[optind]);
			return 1;
		}
	}

	/*  Find maximum column that must be read from each input line.  This
	 *  is determine from the keys columns and data columns requested.
	 *  By knowing the max column, and if the max column needed is less
	 *  than the total number of columns in the input line, we can save
	 *  significant work by spliting only the needed number of fields
	 *  instead of spliting all the fields.  */
	max_col = -1;
	max_col = get_max_src(&key_action, max_col);
	max_col = get_max_src(&data_action_inc_m, max_col);
	
	/*  Storage for data to read when adding a new key to hash.  */
	/*  Determine the maximum number of data elements to store per hash
	 *  element, whether storing input data or processing output data.
	 *  The firstdata[] array is copied to each hash element at creation
	 *  time, so the size if firstdata[] will determine the size of the
	 *  hash elements.  */
	nfirstdata = get_max_dst(&data_action_inc_m,0);
	nfirstdata = get_max_dst(&data_action_out_m,nfirstdata);
	nfirstdata = nfirstdata + 1;
	firstdata  = (double *) calloc (nfirstdata, sizeof(double));

	/*  Initialize hash table  */
	ht = ht_init(nhashslots, max_queue_length ? HT_HISTORY : 0);

	/*  Read file  */
	while (NULL!=fgets(buffer,NSTR,fin)) {

		/*  Remove trailing new line  */
		buffer[strlen(buffer)-1] = '\0';

		/*  Skip comments  */
		if (is_comment(buffer))  continue;

		/*  Read fields from input line  */
		nfield = get_fields (buffer, fptr, max_col+1, field_sep);
		if (max_col>=nfield) continue;

		/*  Form key from input columns  */
		make_key(keybuff, fptr, &key_action, field_sep, NSTR);


		/*  Lookup key, process entry  */
		if ( 
			(t=ht_findelem (ht,(U_CHAR *)&keybuff ,strlen(keybuff)+1))
			) {

			/*  Get pointer to data for this hash element  */
			rdata = (double *) t->data;

			convert_data_inc (&data_action_inc_m, rdata, fptr, FIRST_FALSE);
	
			/*  Make current key the newest, so data output can be newest
			 *  first  */
			ht_makenewest (ht, t);
			
		/*  No key found, create new entry  */
		} else {

			convert_data_inc (&data_action_inc_m, firstdata, fptr, FIRST_TRUE);

			/*  Store new data with new key  */
			ht_storekey (ht, (U_CHAR *) &keybuff, strlen(keybuff)+1, 
				(U_CHAR *) firstdata, nfirstdata*sizeof(firstdata[0]));

		}  /* add new entry  */

		/*  Too many keys - print and remove oldest  */
		if (max_queue_length && ht_getcount(ht)>max_queue_length) {
			t = ht_getoldest(ht);
			convert_data_out (&data_action_out_m, (double *) t->data);
			ht_freeelem(ht,t);
		}

	}   /*  read file  */
		
	fclose(fin);

	/*  If no elements, then suppress sort  */
	if (ht_getcount(ht)==0) {
		use_sort = 0;
		max_output = 0;
	}


	/*  When only printing the first N elements, a full sort is
	 *  relatively inefficient.  Instead, use a simple bubble sort if
	 *  first N element is less then 1/3 the total number.
	 *  This value of 1/3 is a very rough efficiany estimate, feel free
	 *  to change this if you have a better idea.
	 */
	if (use_sort && max_output && max_output<ht_getcount(ht)/3)
		use_top_sort = 1;

	/*  Do final output conversion on all stored data  */
	ht_initwalk(ht);
	while ( (t = ht_getnext(ht)) ) {
		convert_data_out (&data_action_out_m, (double *) t->data);
	}

	/*  Print results in sorted order of data, top max_output only  */
	if (use_top_sort) {

		/*  Allocate index storage for number of top data  */
		index = (helem_t **) calloc(max_output, sizeof(helem_t *));
		if (NULL==index) {
			printf ("ERROR:  Cannot allocate memory for top_sort index\n");
			return 1;
		}

		/*  Place the first max_output elements into index  */
		ht_initwalk(ht);
		for (i=0;i<max_output;i++) {
			index[i] = ht_getnext(ht);
		}

		/*  Read through rest of list comparing with max_index  */
		max_index = get_max_index(index,max_output,dcmp);
		while ( (t = ht_getnext(ht)) ) {
			if (dcmp(index+max_index,&t)>0) {
				index[max_index] = t;
				max_index = get_max_index(index,max_output,dcmp);
			}
		}

		/*  Now have pointers to top max_output elements 
		 *  next sort and print
		 */
		qsort (index, max_output, sizeof(helem_t*), dcmp);
		for (i=0;i<max_output;i++) {
			t = index[i];
			print_key (t);
		}
		free (index);


	/*  Perform a full sort  */
	} else if (use_sort) {
		ht_initwalk(ht);
		index = (helem_t **) calloc (ht_getcount(ht), sizeof(helem_t *));
		if (NULL==index) {
			printf ("ERROR:  Cannot allocate memory for sort index\n");
			return 1;
		}
		i=ht_getcount(ht);
		do {
			i--;
			index[i] = ht_getnext(ht);
		} while (i);
		qsort (index, ht_getcount(ht), sizeof(helem_t*), dcmp);
		for (i=0;i<ht_getcount(ht);i++) {
			t = index[i];
			if (max_output && i>=max_output)
				break;
			print_key (t);
		}
		free (index);

	/*  Print results without sorting  */
	} else {
	
		/*  Write results  */
		ht_initwalk (ht);
		while ((t=ht_getnext(ht))) {
			/*  Print key and data  */
			print_key (t);
		}
	
	}

	/*  Print debug information if asked  */
	if (debug_m) ht_debuginfo(ht);
		
	/*  Don't bother freeing table, it takes alot of time cause we free each node individually,
	 *  just let OS free it up, it goes *much* faster
	 */
	/*  Free table  */
	/*
	ht_free (ht);
	*/

	return 0;
}
	


/*
------------------------------------------------------------------------
Local functions
------------------------------------------------------------------------
*/
int is_delimiter(int c, char *delimiter) {
	return (NULL!=strchr(delimiter,c));
}

int is_whitespace(int c) {
	return (NULL!=strchr(" \t\n\r",c));
}

int is_comment(char *s) {
	while (*s && is_whitespace(*s)) s++;
	return (*s==comment_char_m[0]);
}

/*  Return array of pointers fptr to fields in string  */
int get_fields (char *s, char **fptr, int max_field, int field_sep) {
	int nfield=0;

	/*  Fields separated by one or more blanks - leading blanks ignored  */
	if (field_sep==' ') {

		/*  Skip leading whitespace  */
		while (*s==' ' || *s=='\t') s++;
		/*  Find tokens  */
		while (*s && nfield<max_field) {
			fptr[nfield++] = s;
			/*  Find first whitespace  */
			while (*s!=' ' && *s && *s!='\t')  s++;
			/*  Make first white space string terminator */
			if (*s)  *s++ = '\0';
			/*  Find last whitespace */
			while (*s==' ' || *s=='\t')  s++;
		}

	/*  Fields separated by single field separator  */
	} else {
		
		/*  Find tokens  */
		while (*s && nfield<max_field) {
			fptr[nfield++] = s;
			/*  Find next field separator  */
			while (*s && *s!=field_sep)  s++;
			/*  Make field separator string terminator */
			if (*s)  *s++ = '\0';
		}

	}

	return nfield;
}


/*  Return list_col[] index which is result of 'action' on
 *  output data column 'col'
 */ 
int find_action_col (action_t *proc, int col, int action) {
	int i;
	for (i=0;i<proc->n;i++) {
		if (proc->src[i]==col && proc->action[i]==action) 
			return i;
	}
	return FOUND_FALSE;
}


/*  Print double in integer format if integer  */
void prformat(double d) {
	if (rint(d)==d)  printf (" %.0f", d);
	else             printf (" %e",   d);
}


/*  Return 0,1 if string non-numeric,numeric */
int isnumber (char *str) {
	while (*str) {
		if (*str<'0'  || *str>'9')
			return 0;
		str++;
	}
	return 1;
}


/*  Print key and data for this hash element  */
void print_key (helem_t *t) {
	int    idata;
	double *rdata;

	/*  No element - print blank line  */
	if (t==NULL) {
		printf ("\n");
		return;
	}

	/*  Print key  */
	printf ("%s", t->key);
	/*  Print requested fields  */
	rdata = (double *) t->data;
	for (idata=0;idata<data_action_out_m.nout;idata++) {
		prformat (rdata[idata]);
	}
	printf ("\n");
}


void convert_data_out(action_t *out, double *rdata) {
	int i;
	/*  Zero out->buffer  */
	for (i=0;i<out->nout;i++) out->buf[i] = 0.0;
	for (i=0;i<out->n;i++) {
			switch(out->action[i]) {
				case 'm':
				case 'x':
				case 'f':
				case 'l':
				case 's':
				case 'n':
						out->buf[out->dst[i]] = rdata[out->src[i]];
						break;
				case 'a':
						if (rdata[0]==0.0) {
							out->buf[out->dst[i]] = 0;
						} else {
							out->buf[out->dst[i]] = rdata[out->src[i]]/rdata[0];
						}
						break;
				case 'z':
						if (out->buf[out->src[i]]==0) 
							rdata[out->dst[i]] = 0;
						else
							rdata[out->dst[i]] = 1;
						break;
				case '+':
						out->buf[out->dst[i]] += rdata[out->src[i]];
						break;
				case '-':
						out->buf[out->dst[i]] -= rdata[out->src[i]];
						break;
				case '*':
						out->buf[out->dst[i]] *= rdata[out->src[i]];
						break;
				case '/':
						if (rdata[out->src[i]]==0.0) {
							out->buf[out->dst[i]] = 0.0;
						} else {
							out->buf[out->dst[i]] /= rdata[out->src[i]];
						}
						break;
			}
	}
	/*  Copy temp out->buffer back to data element  */
	for (i=0;i<out->nout;i++) rdata[i] = out->buf[i];
}


void walk_history (htable_t *ht) {
		  /*test*/
		  helem_t *e = ht_getoldest(ht);
		  while (e) {
			  print_key (e);
			  e = (helem_t *) e->newer;
		  }
}



/*  Callback function called by qsort() to order output records  */
int dcmp (const void *aptr, const void *bptr) {
	helem_t *a = * ( (helem_t **) aptr );
	helem_t *b = * ( (helem_t **) bptr );
	double *ad = (double *) a->data;
	double *bd = (double *) b->data;
	double diff;
	int i,j,idiff;

	for (i=0;i<sort_action_m.n;i++) {
		j = sort_action_m.src[i];
		diff = ad[j] - bd[j];
		if (diff==0.0) 
			continue;
		idiff = diff>0 ? 1 : -1;
		if (sort_action_m.action[i]=='r')
			return -idiff;
		else
			return idiff;
	}
	return 0;
}


/*  
Parse data column argument 

arg               string being parsed for data column and actions
def_char          default action code

Note that legal action code 'n' will result in column number -1.
Need to take precautions.

*/
void parse_column (char *arg, int def_char, action_t *t) {
	char *ptr;
	char lastchr;
	int  col;
	t->n=0;
	ptr=strtok(arg,",");
	while (ptr) {
		/*  Print error if too many data columns  */
		if (t->n>MAX_COL) {
			fprintf (stderr,
				"ERROR:  To many data columns specified.  Max allowed is %d\n", 
				MAX_COL);
			exit (1);
		}
		/*  Get trailing character if a,m,n,s,x  */
		lastchr = ptr[strlen(ptr)-1];
		if (!strchr("0123456789", lastchr)) {
			t->action[t->n] = lastchr;
			ptr[strlen(ptr)-1] = '\0';
		} else {
			t->action[t->n] = def_char;
		}
		/*  Read column number for this data item  */
		col = atoi(ptr)-1;
		t->src[(t->n)  ] = col;
		t->dst[(t->n)++] = col;
		ptr=strtok(NULL,",");
	}
}

int copychars(char *p, char *buf, int nbuf, int is_digits) {
	int n = 0;
	if (is_digits)
		while (p[n] && p[n]>='0' && p[n]<='9' && n<nbuf-1) {
			buf[n] = p[n];
			n++;
		}
	else 
		while (p[n] && (p[n]<'0' || p[n]>'9') && n<nbuf-1) {
			buf[n] = p[n];
			n++;
		}
	buf[n]='\0';
	return n;
}


/*  Strings argument is formatted like 1s1,2s2,...  */
void parse_manual (char *arg, int def_char, action_t *t) {
	char *ptr;
	char buf[100];  /*  Ridiculously oversized option char buffer  */
	int  nbuf = 100;
	int  nread1,nread2;
	int  i;

	t->n=0;
	ptr=strtok(arg,",");
	while (ptr) {
		/*  Print error if too many data columns  */
		if (t->n>MAX_COL) {
			fprintf (stderr,
				"ERROR:  To many data columns specified.  Max allowed is %d\n", 
				MAX_COL);
			exit (1);
		}
		/*  Get first digits  */
		nread1 = copychars (ptr, buf, nbuf, 1);
		t->src[t->n] = atoi(buf)-1;
		/*  Read symbol if present */
		nread2 = copychars (ptr+nread1, buf, nbuf, 0);
		if (buf[0]=='\0')
			t->action[t->n] = 's';
		else
			t->action[t->n] = buf[0];
		/*  Read last digits if present  */
		copychars (ptr+nread1+nread2, buf, nbuf, 1);
		if (buf[0]=='\0') 
			t->dst[t->n] = t->src[t->n];
		else
			t->dst[t->n] = atoi(buf)-1;
		if (t->src[t->n]<0 || t->dst[t->n]<0) {
			printf ("ERROR:  Cannot use zero for a manual column\n");
			exit(1);
		}
		/*  Increment number of stored actions  */
		t->n++;
		ptr=strtok(NULL,",");
	}

	/*  Number of columns to which output conversion writes  */
	t->nout = -1;
	for (i=0;i<t->n;i++) {
		if ( t->nout < t->dst[i] )  
			t->nout = t->dst[i];
	}
	t->nout++;
}


/*  Search list of pointers for smallest element  */
int get_max_index (
		helem_t **index, 
		int n,
		int (*cmp)(const void *, const void *))
{
	int max = 0;
	int j;
	for (j=1;j<n;j++) {
		if ( cmp( index+max, index+j )<0 ) {
			max = j;
		}
	}
	return max;
}


/*  Apply action and input value to stored value  */
void convert_data_inc(action_t *inc, double *dout, char **fptr, int firsttime) {
	int i;
	double *dptr, val;
	for (i=0;i<inc->n;i++) {
		if (inc->action[i]!='n') {
			val  = atof(fptr[inc->src[i]]);
		}
		dptr = dout + inc->dst[i];
		switch (inc->action[i]) {
			/*  Store sum  */
			case 'a':
			case 's':
			case '+':
				if (firsttime) {
					*dptr  = val;
				} else {
					*dptr += val;
				}
				break;
			/*  Subtract  */
			case '-':
				if (firsttime) {
					*dptr  = -val;
				} else {
					*dptr -= val;
				}
				break;
			/*  Multiply */
			case '*':
				if (firsttime) {
					*dptr  = -val;
				} else {
					*dptr *= val;
				}
				break;
			/*  Division */
			case '/':
				if (firsttime) {
					*dptr  = -val;
				} else {
					if (val==0.0)
						*dptr = 0.0;
					else 
						*dptr /= val;
				}
				break;
			/*  Zero */
			case 'z':
				if (val!=0.0) val = 1.0;
				if (firsttime) {
					*dptr  = val;
				} else {
					*dptr += val;
				}
				break;
			/*  Store variance (sum of squares)  */
			case 'd':
			case 'e':
				if (firsttime) {
					*dptr  = val*val; } else {
					*dptr += val*val;
				}
				break;
			/*  Do nothing, this is handled automatically  */
			case 'n':
				if (firsttime) {
					*dptr = 1.0;
				} else {
					*dptr += 1.0;
				}
				break;
			/*  Store minimum value  */
			case 'm':
				if (firsttime) {
					*dptr = val;
				} else {
					if (val<*dptr) *dptr = val;
				}
				break;
			/*  Store maximum value  */
			case 'x':
				if (firsttime) {
					*dptr = val;
				} else {
					if (val>*dptr) *dptr = val;
				}
				break;
			/*  Store first occurance  */
			case 'f':
				if (firsttime) *dptr = val;
				break;
			/*  Store last occurance  */
			case 'l':
				*dptr = val;
				break;
		}
	}
}


/*  Read configuration for output conversion to obtain data conversions
 *  on input  */
void get_inc_from_out (action_t *out, action_t *inc) {
	int i;

	/*  By default always count number of occurances for each key  */
	inc->n = 0;
	inc->src   [ inc->n   ] =   0;
	inc->dst   [ inc->n   ] =   0;
	inc->action[(inc->n)++] = 'n';

	for (i=0;i<out->n;i++) {

		/*  See if combination of out_col[i],out_action[i] is not
		 *  already present in inc->src,inc->action.  If so, then we
		 *  don't need to store it again.  This can happen because some
		 *  out_action[] values (for example 'd' or 'e') will also
		 *  add an 's' value.  */
		if (FOUND_FALSE != find_action_col(inc, out->src[i], out->action[i]) )
			continue;

		/*  Print error if too many out items */
		if (inc->n==MAX_COL)  {
			fprintf (stderr, 
				"ERROR:  To many out actions specified.  Max allowed is %d\n", MAX_COL);
			exit (1);
		}

		/*  Save out column action in incess  */
		switch (out->action[i]) {
			/*  Field counts happens automatically  */
			case 'n':
				out->src[i] = 0;
				out->dst[i] = i;
				break;
			/*  'Simple' incesses */
			case 'm':  /*  Minimum  */
			case 'x':  /*  Maximum  */
			case 'a':  /*  Average  */
			case 's':  /*  Sum      */
			case 'f':  /*  First    */
			case 'l':  /*  Last     */
			case 'z':  /*  Zero/One */
			case '+':  /*  Add      */
			case '-':  /*  Subtract */
			case '*':  /*  Multiply */
			case '/':  /*  Divide   */
				inc->action[inc->n] = out->action[i];
				inc->src   [inc->n] = out->src[i];
				inc->dst   [inc->n] = inc->n;
				out->src   [i     ] = inc->dst[inc->n];
				out->dst   [i     ] = i;
				inc->n++;
				break;
		}
	}

	/*  Number of columns to which input conversion writes  */
	out->nout = -1;
	for (i=0;i<out->n;i++) {
		if ( out->nout < out->dst[i] )  
			out->nout = out->dst[i];
	}
	out->nout++;

	/*  Number of columns to which output conversion writes  */
	inc->nout = -1;
	for (i=0;i<inc->n;i++) {
		if ( inc->nout < inc->dst[i] )  
			inc->nout = inc->dst[i];
	}
	inc->nout++;
}


/*  Find maximum position stored in src arrays in action_t  */
int get_max_src(action_t *a, int max) {
	int i;
	for (i=0;i<a->n;i++)
		if (max < a->src[i])  max = a->src[i];
	return max;
}


/*  Find maximum position stored in dst arrays in action_t  */
int get_max_dst(action_t *a, int max) {
	int i;
	for (i=0;i<a->n;i++)
		if (max < a->dst[i])  max = a->dst[i];
	return max;
}


/*  Make records key from appropriate input columns  */
void make_key(char *keybuff, char **fptr, action_t *key, int field_sep, int nlen) {
	int ikey;
	char sep_str[2] = "\0\0";
	sep_str[0] = field_sep;
	keybuff[0] = '\0';
	nlen--;
	if (key->n) {
		for (ikey=0;ikey<key->n;ikey++) {
			/*  Prepend field separator except for first field  */
			if (ikey!=0) {
				strncat (keybuff, sep_str, nlen);
				nlen -= 1;
			}
			strncat (keybuff, fptr[key->src[ikey]],nlen);
			nlen -= strlen(fptr[key->src[ikey]]);
		}
	}
}

void debug_action(char *msg, action_t *a) {
	int i;
	printf ("actionname action_p len: (%s) (%p) (%d)\n", msg, a, a->n);
	for (i=0;i<a->n;i++) 
		printf ("  src dst action %2d %2d %c\n", a->src[i], a->dst[i], a->action[i]);
}

void Usage(void) {
		char **ptr = Usage_m;
		printf ("%s\n", *ptr);
		while (*ptr) {
			printf ("%s\n", *(ptr++));
		}
}

